/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
/*
 *  FILE        : approx.c
 *  DESCRIPTION : for each A = R|G|B, SVD, Approximate with rank k
 *
 *  HISTORY:
 *    06/23/2008 Song Zhenhua
 *      - Initial version
 */
/* 
 * The main process of approx.c
 * Step-1 for DGESVD/SDD, Initialize U, S, VT, LDA and so on
 * Step-2 DGESVD/SDD
 * Step-3 Verify. Notice this is AA, NOT A 
 * Step-4 Approximation with rank k
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "approx.h"
#include "verify.h"

void SVD(int m, int n, double *A, int lda, double *U, \
                int ldu, double *S, double *VT, int ldvt)
{
    char *str = NULL; /* get ENV VAL, DGESVD or DGESDD */
    double *WORK = NULL;
    double size = 0.0;
    int lwork, info = 0;

    str = getenv("SVDFUNC");
    if ((NULL != str) && (0 == strcmp(str, "DGESVD")))
    {
        /* By dgesvd */
#ifdef _DEBUG
        printf("\t\t*** CALL DGESVD **** \n");
#endif
        /* Get maxwork */
        lwork = -1;
        dgesvd_("A", "A", &m, &n, A, &lda, S, U, &ldu, VT, &ldvt, \
                        &size, &lwork, &info);
        lwork = (int)size;
#ifdef _DEBUG
        printf("\tLWORK = %d\n", lwork);
#endif
        WORK = malloc(lwork * sizeof(double));
        if (WORK == NULL)
        {
            printf(" Malloc WORK for DGESVD failed!\n");
            exit(1);
        }

        /* CALL ROUTINE */
        dgesvd_("A", "A", &m, &n, A, &lda, S, U, &ldu, VT, &ldvt, \
                        WORK, &lwork, &info);
    }
    else
    {
        /* By degsdd, default */
#ifdef _DEBUG
        printf("\t\t*** CALL DGESDD **** \n");
#endif
        int *iwork = NULL;
        iwork = malloc(8 * MIN(m, n) * sizeof(int));
        if (iwork == NULL)
        {
            printf(" Malloc iwork for DGESDD failed!\n");
            exit(1);
        }

        /* Get maxwork */
        lwork = -1;
        dgesdd_("A", &m, &n, A, &lda, S, U, &ldu, VT, &ldvt, \
                        &size, &lwork, iwork, &info);
        lwork = (int)size;
#ifdef _DEBUG
        printf("\tLWORK = %d\n", lwork);
#endif
        WORK = malloc(lwork * sizeof(double));
        if (WORK == NULL)
        {
            printf(" Malloc WORK for DGESDD failed!\n");
            exit(1);
        }

        /* CALL ROUTINE */
        dgesdd_("A", &m, &n, A, &lda, S, U, &ldu, VT, &ldvt, \
                        WORK, &lwork, iwork, &info);
        /* free iwork */
        free(iwork);
        iwork = NULL;
    }

    /* free WORK */
    free(WORK);
    WORK = NULL;

    if(info != 0)
    {
        printf("svd execute error info : %d\n", info);
        exit(1);
    }

    return;
}

/* Approximation Using the first k cols of U and k rows of VT */
void k_approx(int m, int n, double *A, int lda, double *U, \
                int ldu, double *S, double *VT, int ldvt, int k)
{
#ifdef _DEBUG
    printf(" %s\n", __FUNCTION__);
#endif
    int i, j;
#ifdef _DEBUG
    printf(" Useing dgemm() ...\n");
#endif
    double *MS, *C;     /* convert Vector S to Matrix MS, C store the result of U * MS */
    MS = C = NULL;
    double alpha, beta;

    /* Initialize MS and C*/
    MS = malloc(k * k * sizeof(double));
    if (MS == NULL)
    {
        printf(" Malloc Matrix S failed!\n");
        exit(1);
    }
    for (i = 0; i < k; i++)
    {
        for (j = 0; j < k; j++)
        {
            MS[i + j * k] = (i == j) ? S[i] : 0.0;
        }
    }
    C = malloc(m * k * sizeof(double));
    if (C == NULL)
    {
        printf(" Malloc Matrix C failed!\n");
        free(MS);
        MS = NULL;
        exit(1);
    }

    /* U * MS -> C */
    alpha = 1.0;
    beta = 0.0;
    dgemm("N", "N", &ldu, &k, &k, &alpha, U, &ldu, MS, &k, &beta, C, &m);
    free(MS);
    MS = NULL;

    /* C * VT -> A */
    dgemm("N", "N", &m, &n, &k, &alpha, C, &m, VT, &ldvt, &beta, A, &lda);
    free(C);
    C = NULL;
    return;
}

void Approx(double *A, unsigned int m, unsigned int n, unsigned int k)
{
    double *AA, *U, *S, *VT;       /* U, S, VT for svd. AA is COPY of A */
    AA = U = S = VT = NULL;
    unsigned int lda, ldu, ldvt;
    unsigned int min;
    int flag = 0;                  /* flag of exit(error) or return(no error) */
    unsigned i;

    /* initialize */
    lda = ldu = m;
    ldvt = n;
    min = MIN(m, n);

    AA = malloc(lda * n * sizeof(double));
    U = malloc(ldu * m * sizeof(double));
    S = malloc(min * sizeof(double));
    VT = malloc(ldvt * n * sizeof(double));

    if ((NULL == AA) || (NULL == U) || (NULL == S) || (NULL == VT))
    {
        printf(" Malloc matrix for svd failed!\n");
        flag = 1;
        goto ret_approx;
    }

    for (i = 0; i < m * n; i++)
    {
        AA[i] = A[i];
    }

    /* SVD */
#ifdef _DEBUG
    /* Timing */
    struct timeval begin, end;
    double time;
    TDS_RB();
#endif
    SVD(m, n, A, lda, U, ldu, S, VT, ldvt);
#ifdef _DEBUG
    TDS_RE();
    time = TDS_RGET();
    printf("\t\t *** SVD Time : %4.8f ***\n", time);

    /* Verify. Notice this is AA, NOT A */
    TDS_RB();
#endif    
    verify(m, n, AA, lda, U, ldu, S, VT, ldvt);
#ifdef _DEBUG
    TDS_RE();
    time = TDS_RGET();
    printf("\t\t *** Verify Time : %4.8f ***\n", time);

    /* Approximation with rank k */
    TDS_RB();
#endif    
    k_approx(m, n, A, lda, U, ldu, S, VT, ldvt, k);
#ifdef _DEBUG
    TDS_RE();
    time = TDS_RGET();
    printf("\t\t *** K-approx Time : %4.8f ***\n", time);
#endif
    /* return */
ret_approx :
    if(NULL != AA)
    {
        free(AA);
        AA = NULL;
    }
    if(NULL != U)
    {
        free(U);
        U = NULL;
    }
    if(NULL != S)
    {
        free(S);
        S = NULL;
    }
    if(NULL != VT)
    {
        free(VT);
        VT = NULL;
    }
    if (flag)
    {
        exit(1);
    }
    return;
}
