/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
/*
 *  FILE        : main.c
 *  DESCRIPTION : Digital Image Approximation(DIA) using Singular Value
 *                Decomposition(SVD)
 *  HISTORY:
 *    06/23/2008 Song Zhenhua
 *      - Initial version
 */

/*
 * We can get R(red), G(green) and B(blue) color matrix(m * n) from a picture, 
 *and for each A = R|G|B, A = U * S * VT, namely SVD. If we set :
 *   Ui is the ith col of U, and
 *   Si is the ith element of S, and 
 *   VTi is the ith row of VT, then 
 * A = sum of Ui * Si * VTi, which i from 1 to MIN(m, n);
 * Because elements of S are arranged degressively, we can select a rank k and 
 * approximate A using :
 *    A' = sum of Ui * Si * VTi, which i from 1 to k.
 * We can save space form m*n to k*(m+n+1) in this way, and transmit it more 
 * quickly.
 * As a sample, svd can only process uncompressed bitmap picture file.
*/
/* 
 * The main process of main.c
 * Step-1 Read source file and get some valueable datas
 *        including height and width of source file, and so on
 * Step-2 Malloc space for R, G, B
 * Step-3 Write head parts to target file 
 * Step-4 Write to R, G, B from source file
 * Step-5 input k, SVD with rank k, and verify
 * Step-6 Write bits data to target file, and process over
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "bmp.h"
#include "approx.h"
#include "verify.h"

/* File Pointer of Bitmap files of source and target 
 * Digital Image Approximation */
FILE *fpin, *fpout;

int main(int argc, char *argv[])
{
    unsigned char head[BM_HEAD_SIZE]; /* pointer to head of bitmap file */
    unsigned int start;               /* start of real bit data */
    unsigned int height;              /* height of bitmap file */
    unsigned int width;               /* width of bitmap file */
    unsigned int i, j, n;             /* for loop or tmp use */
    unsigned int k = 0;               /* rank for Approximation with SVD */
    unsigned char ch;                 /* char read from source bitmap file */
    double *R, *G, *B;                /* color matrix of R, G, B */
    R = G = B = NULL;

    /* parse parameters of command line */
    if (argc < 2)
    {
        printf(" Please input a source bitmap file\n");
        return 0;
    }
    else if (argc < 3)
    {
        printf(" Please input a target bitmap file\n");
        return 0;
    }
    else if (argc > 3)
    {
        printf(" Too many parameters\n");
        return 0;
    }
        /* Check if source and target files is .bmp. FIXME */

    /* Read source file and get some valueable datas */
    fpin = fopen(argv[1], "rb");
    if(NULL == fpin)
    {
        printf(" Can't open source file\n");
        return 0;
    }

    fread(head, BM_HEAD_SIZE, 1, fpin);

    /* Test Bitmap Type */
    if ((head[0] != 'B') || (head[1] != 'M'))
    {
        printf(" svd ONLY Handle Windows version, NOT OS/2 or Others!\n");
        return 1;
    }

    /* Test Bit Counts and Compression */
    if (NUM(BI_BCC) != 24)
    {
        printf(" svd ONLY Handle NON-Compressed 24-bit Bitmap!\n");
        return 1;
    }    

    start = NUM(BF_OFFBITS);
    width = NUM(BI_WIDTH);
    height = NUM(BI_HEIGHT);
#ifdef _DEBUG
    printf(" Width bits of %s is %u\n", argv[1], width);
#endif
    /* biWidth' is the smallest integer which is no less than biWidth 
     * and indivisible by 4 */
    if ((width % 4) != 0)
    {
        width = ((width / 4) + 1) * 4;
    }
#ifdef _DEBUG
    printf(" Source file %s is %u * %u(width * height), image datas start at %u\n", \
                         argv[1], width, height, start);
#endif

    /* Malloc space for R, G, B */
    R = malloc(width * height * sizeof(double));
    G = malloc(width * height * sizeof(double));
    B = malloc(width * height * sizeof(double));
    if ((NULL == R) || (NULL == G) || (NULL == B))
    {
        printf(" Malloc color matrix failed!\n");
        goto ret_main;
    }

    /* Write head parts to target file */
    fseek(fpin, SEEK_SET, 0);
    fpout = fopen(argv[2], "wb");
    if(NULL == fpout)
    {
        printf(" Can't open target file\n");
        goto ret_main;
    }
    for (i = 0; i < start; i++)
    {
        fputc(fgetc(fpin), fpout);
    }

    /* Write to R, G, B from source file */
    for (i = 0; i < width * height * 3; i++)
    {
        ch = fgetc(fpin);
        j = i % 3;
        n = i / 3;

        /* remainder is 0 : blue, 1 : green, 2 : red */
        switch (j)
        {
            case 1 :
                G[n] = (double)ch;
                break;
            case 2 :
                R[n] = (double)ch;
                break;
            default :
                B[n] = (double)ch;
                break;
        }
    }

    /* input rank k for Approximation */
    printf("\tBitmap file %s is %u * %u(width * height)\n\tPlease input rank for Approximation image : ", \
                         argv[1], width, height);
inputk :
    scanf("%u", &k);
    if (k > MIN(width, height))
    {
        printf(" k should be less than %u, input again please\n", MIN(width, height));
        goto inputk;
    }

    /* Computing Saved Space */
    double save;
    save = 1 - 1.0 * k * (width + height + 1) / (width * height);
    printf(" You can save space by %3.2f%% !\n", save * 100);

    /* Digital Image Approximation by SVD with rank k, and verify */
#ifdef _DEBUG
    printf(" Now begin Approximation by SVD with rank %u\n", k);
    printf(" Approx R ...\n");

    /* Timing */
    struct timeval begin, end;
    double time;
    TDS_RB();
#endif
    Approx(R, height, width, k);
#ifdef _DEBUG
    printf(" Approx G ...\n");
#endif
    Approx(G, height, width, k);
#ifdef _DEBUG
    printf(" Approx B ...\n");
#endif
    Approx(B, height, width, k);

    /* End of Approximation, Write bits data to target file */
#ifdef _DEBUG
    TDS_RE();
    time = TDS_RGET();
    printf("\n\t *** Approximate Time : %4.8f ***\n\n Now Write bits data to %s ...\n", \
                time, argv[2]);
#endif
    for (i = 0; i < width * height * 3; i++)
    {
        j = i % 3;
        n = i / 3;

        switch (j)
        {
            case 1 :
                ch = (unsigned int)G[n];
                break;
            case 2 :
                ch = (unsigned int)R[n];
                break;
            default :
                ch = (unsigned int)B[n];
                break;
        }

        fputc(ch, fpout);
    }

    /* Approximation Over */
#ifdef _DEBUG
    printf(" %s OK, return\n", argv[2]);
#endif
ret_main :
    if (NULL != fpin)
    {
        fclose(fpin);
    }
    if (NULL != fpout)
    {
        fclose(fpout);
    }
    if(NULL != R)
    {
        free(R);
        R = NULL;
    }
    if(NULL != G)
    {
        free(G);
        G = NULL;
    }
    if(NULL != B)
    {
        free(B);
        B = NULL;
    }
    return 0;
}
