/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
/*
 *  FILE        : verify.c
 *  DESCRIPTION : verify results of dgesvd/sdd
 *
 *  HISTORY:
 *    06/23/2008 Song Zhenhua
 *      - Initial version
 */
/* 
 * The main process of verify.c
 * Step-1 initialize, get dgemm type and 
 * Step-2 Test 1 : DBDT01 :  | A - U diag(S) VT | / ( |A| max(M,N) ulp
 * Step-3 Test 2 : DORT01 : | I - U'U | / ( M ulp ) 
 * Step-4 Test 3 : DORT01 : | I - VT VT' | / ( N ulp )
 * Step-5 Test 4 : S contains MNMIN nonnegative values in decreasing order
*/

#include <stdlib.h>
#include <stdio.h>

#include "verify.h"

/* Simplified DLANSY, norm1(A), NORM = '1', 'O' or 'o'
 * DOUBLE PRECISION FUNCTION DLANSY( NORM, UPLO, N, A, LDA, WORK ) 
 * 
 * Need PWORK : kx + n
 * */
double dlansy(char *uplo, int *n, double *p_A, int *lda, double *p_work, int kx)
{
#ifdef _DEBUG
    printf(" %s()\n", __FUNCTION__);
#endif
    int i, j;
    double value, sum, absa;

    value = 0.0;

    /* Upper */
    if (*uplo == 'U')
    {
        for (j = 0; j < *n; j++)
        {
            sum = 0.0;
            for (i = 0; i < j; i++)
            {
                absa = ABS(p_A[i + j * (*lda)]);
                sum = sum + absa;
                /* Notice kx!! */
                p_work[kx + i] = p_work[kx + i] + absa;
            }
            p_work[kx + j] = sum + ABS(p_A[j + j*(*lda)]);
        }
        for (i = 0; i < *n; i++)
        {
            value = MAX(value, p_work[kx + i]);
        }
    }
    /* Lower */
    else
    {
        for (i = 0; i < *n; i++)
        {
            p_work[kx + i] = 0.0;
        }
        for (j = 0; j < *n; j++)
        {
            sum = p_work[kx + j] + ABS(p_A[j + j * (*lda)]);
            for (i = j + 1; i < *n; i++)
            {
                absa = ABS(p_A[i + j * (*lda)]);
                sum = sum + absa;
                /* Notice kx!! */
                p_work[kx + i] = p_work[kx + i] + absa;
            }
            value = MAX(value, sum);
        }
    }

    return value;
}

/* Need PWORK : (ldwork + 1) * min = m * n + MIN(m, n) */
void dort01(char *rowcol, int *m, int *n, double *U, int *ldu, \
                double *work, double *resid)
{
#ifdef _DEBUG
    printf(" %s()\n", __FUNCTION__);
#endif
    double eps = dlamch_("P");
    char uplo;
    int k, min, ldwork;
    double alpha, beta;

    if((*m < *n) || ((*m == *n) && (*rowcol == 'R')))
    {
        k = *n;
    }
    else 
    {
        k = *m;
    }

    min = MIN(*m, *n);
    ldwork = MAX(*m, *ldu);

    alpha = 0.0;
    beta = 1.0;
    dlaset_("Upper", &min, &min, &alpha, &beta, work, &ldwork);

    alpha = 1.0;
    beta = -1.0;
    if (*rowcol == 'C')
    {
        /* I - U'U */
        dgemm("T", "N", m, m, ldu, &alpha, U, ldu, U, ldu, &beta, work, &ldwork);
    }
    else
    {
        /* I - VT VT' */
        dgemm("N", "T", ldu, ldu, n, &alpha, U, ldu, U, ldu, &beta, work, &ldwork);
    }

    uplo = 'U';
    *resid = dlansy(&uplo, &min, work, &ldwork, work, (ldwork * min + 1));
    *resid = ((*resid) / k) / eps;

    return;
}

/* DBDT01. in DDRVBD, E: (workspace) DOUBLE PRECISION array, dimension (max(min(MM,NN))).
 * Need WORK : m * n
 * */
void dbdt01(int *m, int *n, double *p_A, int *lda, double *p_U, int *ldu, double *p_S, double *p_VT, \
            int *ldvt, double *p_work, double *resid)
{
#ifdef _DEBUG
    printf(" %s()\n", __FUNCTION__);
#endif
    int i, j, min;
    double anorm, eps, tmp, alpha, beta;
    char norm;

    /* Compute A - Q * B * P' one column at a time. */
    *resid = 0.0;
    min = MIN(*m, *n);

    /* US -> PWORK */
    for (j = 0; j < min; j++)
    {
        for (i = 0; i < (*ldu); i++)
        {
            p_work[i + j * (*ldu)] = p_U[i + j * (*ldu)] * p_S[j];
        }
    }

    /* AVT' - PWORK -> PWORK */
    alpha = 1.0;
    beta = -1.0;
    dgemm("N", "T", lda, ldvt, n, &alpha, p_A, lda, p_VT, ldvt, &beta, p_work, m);

    /* |A - USV| */
    for (j = 0; j < min; j++)
    {
        tmp = 0.0;
        for (i = 0; i < *m; i++)
        {
            tmp += ABS(p_work[i + j * (*m)]);
        }

        *resid = MAX(*resid, tmp);
    }

    /* Compute norm(A - Q * B * P') / ( n * norm(A) * EPS ) */
    norm = '1';
    anorm = dlange_(&norm, m, n, p_A, lda, p_work);
    eps = dlamch_("P");
    if (anorm <= 0.0)
    {
        if (*resid != 0.0)
        {
            *resid = 1.0 / eps;
        }
    }
    else
    {
        if (anorm >= *resid)
        {
            *resid = (*resid / anorm) / (((double)(*n)) * eps);
        }
        else
        {
            if (anorm < 1.0)
            {
                *resid = (MIN(*resid, (((double)(*n))*anorm)) / anorm) / (((double)(*n)) *eps);
            }
            else
            {
                *resid = MIN((*resid / anorm), ((double)(*n))) / (((double)(*n)) *eps);
            }
        }
    }

    return;
}

/* For Verify, for each DGESVD/DGESDD :
 * (1)    | A - U diag(S) VT | / ( |A| max(M,N) ulp (DBDT01, RESID = norm( A - Q * B * PT ) / ( n * norm(A) * EPS )
 * (2)    | I - U'U | / ( M ulp ) (DORT01)
 * (3)    | I - VT VT' | / ( N ulp ) (DORT01)
 * (4)    S contains MNMIN nonnegative values in decreasing order. (Return 0 if true, 1/ULP if false.) 
 */
void verify(int m, int n, double *A, int lda, double *U, int ldu, double *S, double *VT, int ldvt)
{
#ifdef _DEBUG
    printf(" %s()\n", __FUNCTION__);
#endif
    int i, j, lwork, min;
    char rowcol, *show[4];              /* show : to print test result */
    show[0] = show[1] = show[2] = show[3] = NULL;
    double ulp, ulpinv, result[4], *WORK, PERFTHRE = 30.00;
    WORK = NULL;

    for (i = 0; i < 4; i++)
    {
        show[i] = malloc(FILELINE * sizeof(char));
        if (show[i] == NULL)
        {
            printf(" malloc show[%d] error\n", i);
            for (j = 0; j < i; j++)
            {
                free(show[i]);
                show[i] = NULL;
            }
        }
    }

    /* malloc WORK space */
    lwork = 3 * m * n;
    WORK = malloc(lwork * sizeof(double));
    if (WORK == NULL)
    {
        printf(" malloc WORK error\n");
        exit(1);
    }

    /* Test 1 : DBDT01 :  | A - U diag(S) VT | / ( |A| max(M,N) ulp */
    dbdt01(&m, &n, A, &lda, U, &ldu, S, VT, &ldvt, WORK, &result[0]);
    snprintf(show[0], FILELINE, "Test 1 :     DBDT01 : | A - U diag(S) VT | / ( |A| max(M,N) ulp. . RESULT : %4.8f\n", \
                result[0]);

    /* Test 2 : DORT01 : | I - U'U | / ( M ulp ) */
    rowcol = 'C';
    dort01(&rowcol, &m, &m, U, &ldu, WORK, &result[1]);
    snprintf(show[1], FILELINE, "Test 2 :         DORT01 : | I - U'U | / ( M ulp ).               . RESULT : %4.8f\n", \
                result[1]);

    /* Test 3 : DORT01 : | I - VT VT' | / ( N ulp ) */
    rowcol = 'R';
    dort01(&rowcol, &n, &n, VT, &ldvt, WORK, &result[2]);
    snprintf(show[2], FILELINE, "Test 3 :         DORT01 : | I - VT VT' | / ( N ulp ).            . RESULT : %4.8f\n", \
                result[2]);

    /* Test 4 : S contains MNMIN nonnegative values in decreasing order */
    ulp = dlamch_("P");
    ulpinv = 1.0 / ulp;
    result[3] = 0.0;
    min = MIN(m, n);

    for (i = 0; i < min - 1; i++)
    {
        if (S[i] < S[i+1])
        {
            result[3] = ulpinv;
        }
        if (S[i] < 0.0)
        {
            result[3] = ulpinv;
        }
    }
    if ((min >= 1) && (S[min - 1] < 0))
    {
        result[3] = ulpinv;
    }

    snprintf(show[3], FILELINE, "Test 4 : S contains MNMIN nonnegative values in decreasing order.. RESULT : %4.8f", \
                result[3]);

    /* Collection Garbage */
    if (WORK)
    {
        free(WORK);
        WORK = NULL;
    }
#ifdef _DEBUG
    /* Print 4 test results */
    printf("\n\t\t *** TEST RESULTS : ***\n");
    for (i = 0; i < 4; i++)
    {
        printf("%s", show[i]);
    }
    printf("\n\t\t *** END ***\n");
#endif
    /* Print Failed test results */
    for (i = 0; i < 4; i++)
    {
        if (result[i] > PERFTHRE)
        {
            printf(" * SVD(M = %d, N = %d) fail...\n\t%s\n", m, n, show[i]);
        }
    }
    if (i == 4)
    {
        printf(" * SVD TEST PASS ...\n");
    }

    /* Free shows */ 
    for (i = 0; i < 4; i++)
    {
        free(show[i]);
        show[i] = NULL;
    }

    return;
}
